# Microsoft Azure Linux Agent
#
# Copyright 2018 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# E0401: Unable to import 'lisa' (import-error)
from lisa import (  # pylint: disable=E0401
    Logger,
    Node,
    TestCaseMetadata,
    TestSuite,
    TestSuiteMetadata,
)


@TestSuiteMetadata(area="sample", category="", description="")
class HelloWorld(TestSuite):
    @TestCaseMetadata(description="")
    def main(self, node: Node, log: Logger) -> None:
        log.info(f"Hello world from {node.os.name}!")
